/*
 * Copyright (C) 2006-2011 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
#include <stdbool.h>

/** \file net.h exported symbols from net.c */

/**
 * the buffer size of the sun_path component of struct sockaddr_un
 *
 * While glibc doesn't define \p UNIX_PATH_MAX, it
 * documents it has being limited to 108 bytes.
 */
#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

/* Userland defines for Linux DCCP support. */

#ifndef IPPROTO_DCCP
#define IPPROTO_DCCP 33 /**< IANA assigned value. */
#endif

#ifndef SOCK_DCCP
#define SOCK_DCCP 6 /**< Linux socket type. */
#endif

#ifndef DCCP_SOCKOPT_RX_CCID
/** Per-connection CCID support (set/get the RX CCID, since v2.6.30-rc1). */
#define DCCP_SOCKOPT_RX_CCID 15
#endif

#ifndef SOL_DCCP
#define SOL_DCCP 269 /**< Linux socket level. */
#endif

#ifndef DCCP_SOCKOPT_GET_CUR_MPS
#define DCCP_SOCKOPT_GET_CUR_MPS  5 /**< Max packet size, RFC 4340, 14. */
#endif

#ifndef DCCP_SOCKOPT_AVAILABLE_CCIDS
#define DCCP_SOCKOPT_AVAILABLE_CCIDS 12 /**< List of supported CCIDs. */
#endif

#ifndef DCCP_SOCKOPT_CCID
#define DCCP_SOCKOPT_CCID 13 /**< Sets both TX/RX CCID. */
#endif

#ifndef DCCP_SOCKOPT_TX_CCID
#define DCCP_SOCKOPT_TX_CCID 14 /**< Set/get the TX CCID. */
#endif

/**
 * Flowopts: Transport-layer independent encapsulation of socket options
 *           that need to be registered prior to setting up a connection.
 */
struct flowopts;

extern struct flowopts *flowopt_new(void);
extern void flowopt_add(struct flowopts *fo, int level, int opt,
		const char *name, const void *val, int len);
extern void flowopt_add_bool(struct flowopts *fo, int lev, int opt,
		const char *optname, bool on_or_off);
/** Flowopt shortcut macros */
#define OPT_ADD(fo, lev, opt, val, len)	flowopt_add(fo, lev, opt, #opt, val, len)
#define OPT_ENABLE(fo, lev, opt)	flowopt_add_bool(fo, lev, opt, #opt, 1)
#define OPT_DISABLE(fo, lev, opt)	flowopt_add_bool(fo, lev, opt, #opt, 0)

/**
 * Functions to parse and validate (parts of) URLs.
 */
extern char *parse_cidr(const char *cidr,
			char *addr, ssize_t addrlen, int32_t *netmask);
extern char *parse_url(const char *url,
		       char *host, ssize_t hostlen, int32_t *port);
extern const char *stringify_port(int port, const char *transport);
/**
 * Ensure that string conforms to the IPv4 address format.
 *
 * \param address The address string to check.
 *
 * \return 1 if \a address conforms to the IPv4 address format, else 0.
 */
_static_inline_ bool is_valid_ipv4_address(const char *address)
{
	struct in_addr test_it;

	return inet_pton(AF_INET, address, &test_it) != 0;
}

/**
 * Ensure that string conforms to IPv6 address format.
 *
 * \param address The address string to check.
 *
 * \return 1 if string has a valid IPv6 address syntax, 0 if not.
 * \sa RFC 4291
 */
_static_inline_ bool is_valid_ipv6_address(const char *address)
{
	struct in6_addr test_it;

	return inet_pton(AF_INET6, address, &test_it) != 0;
}

/**
 * Generic socket creation (passive and active sockets).
 */
extern int makesock(unsigned l4type, bool passive,
		    const char *host, uint16_t port_number,
		    struct flowopts *fo);

static inline int para_connect_simple(unsigned l4type,
				      const char *host, uint16_t port)
{
	return makesock(l4type, 0, host, port, NULL);
}

extern struct in_addr extract_v4_addr(const struct sockaddr_storage *ss);

/**
 * Functions to support listening sockets.
 */
/** How many pending connections queue of a listening server will hold. */
#define BACKLOG	10
extern int para_listen(unsigned l4type, uint16_t port, struct flowopts *fo);

static inline int para_listen_simple(unsigned l4type, uint16_t port)
{
	return para_listen(l4type, port, NULL);
}

/** Pretty-printing of IPv4/6 socket addresses */
extern char *local_name(int sockfd);
extern char *remote_name(int sockfd);

/**
 * Determining maximum payload (packet) size
 */
extern int generic_max_transport_msg_size(int sockfd);

int send_bin_buffer(int, const char *, size_t);
int send_buffer(int, const char *);
__printf_2_3 int send_va_buffer(int fd, const char *fmt, ...);

int recv_bin_buffer(int fd, char *buf, size_t size);
int recv_buffer(int fd, char *buf, size_t size);

int para_accept(int fd, fd_set *rfds, void *addr, socklen_t size, int *new_fd);
int create_local_socket(const char *name, struct sockaddr_un *unix_addr,
	mode_t mode);
int connect_local_socket(const char *name);
int recv_cred_buffer(int, char *, size_t);
ssize_t send_cred_buffer(int, char*);

/**
 * Functions and definitions to support \p IPPROTO_DCCP
 */
/** Estimated worst-case length of a DCCP header including options. */
#define DCCP_MAX_HEADER		128
/** Hardcoded maximum number of separate CCID modules compiled into a host. */
#define DCCP_MAX_HOST_CCIDS	20
extern int dccp_available_ccids(uint8_t **ccid_array);
